#
# faq-html.spec
# Example FAQ to HTML CoST conversion specification.
#
# $Revision: 1.3 $
# $Date: 1996/01/16 23:58:14 $
#

require Simple.tcl;

# the 'htmlEscape' command replaces HTML special characters
# with appropriate entity references:
#
#   [htmlEscape "a < b & b < c"] ==> "a &lt; b &amp; b &lt; c"
#
substitution htmlEscape  {
	{<}	{&lt;}
	{>}	{&gt;}
	{&}	{&amp;}
}


# 'nullFilter' is used to suppress output in, e.g., the HEAD.
#
proc nullFilter {text} {
    return ""
}

# 'sdataTranslate' uses a global array to map SDATA entity text
# to the correct output formats:
#
proc sdataTranslate {sdata} {
    global sdataMap
    if [info exists sdataMap($sdata)] {
	return $sdataMap($sdata)
    }
    warning "No translation for SDATA text |$sdata|"
    return $sdata
}

# The DTD only includes a few SDATA entities.
# Note that this keys off the *replacement text*, not the entity name.

global sdataMap
array set sdataMap {
    {[nbsp  ]}  { }
    {[mdash ]}  {--}
}


# Translation parameters:

specification translate {

    {element FAQ} {
	prefix	{<HTML>\n}
	suffix	{</HTML>\n}
	startAction {
	    set qnumber 0 ; 	# current question number
	    set secnum 0 ; 	# current section number
	}
	cdataFilter	htmlEscape
	sdataFilter	sdataTranslate
    }

    {element SECT} {
	startAction { incr secnum }
    }
    {element QNA} {
	startAction { incr qnumber }
    }

    {element Q parent hasatt ID} {
	before	{<H3><A NAME=[query parent attval ID]>}
	after	{</A></H3>}
    }
    {element Q} {
	before  <H3>
	prefix	{<b>Q${qnumber}.</b> }
	after	</H3>
    }
    {element QREF} {
	startAction {
	    set refid [query attval REFID]
	    output "<A HREF=\"#$refid\">"
	    output <B>
	    withNode doctree element QNA withattval ID $refid child element Q {
		output "\"[content]\""
	    }
	    output </B>
	    output </A>
	}
    }

    {element A} {
	prefix	{<P><b>A.</b>  }
    }

    {element P} {
	prefix	{\n<P>}
    }

    {element LIST} {
	prefix	<UL>
	suffix	</UL>
    }
    {element IT} {
	prefix	<LI>
    }

    {element PRE} {
	prefix	{\n<PRE>\n}
	suffix	{\n</PRE>\n}
    }

    {element EXCERPT} {
	prefix	<BLOCKQUOTE>
	suffix	</BLOCKQUOTE>
    }

    {element EMPH} 	{ prefix <EM>  suffix </EM> }
    {element LIT} 	{ prefix <TT>  suffix </TT> }
    {element DFN}	{ prefix <EM>  suffix </EM> }
    {element URL} {
	prefix {<br><a href="[content]">URL:}
	suffix {</a><br>}
    }

    {element HEAD} {
	prefix	<HEAD>
	suffix	</HEAD>
	cdataFilter	nullFilter
	sdataFilter	nullFilter
    }
    {element AUTHOR hasatt EMAIL in HEAD} {
	startAction {
	    output "\n<LINK REV=MADE HREF=\"mailto:[query attval EMAIL]\">\n"
	}
    }

    {element TITLE in HEAD} {
	cdataFilter htmlEscape
	prefix <TITLE>
	suffix </TITLE>
    }
    {element TITLE in SECT} {
	prefix	"\n<H2><A NAME=SEC$secnum>"
	suffix	"</A></H2>\n"
    }

    {element BODY} {
	before <BODY>
	startAction {
	    # copy document title as first-level heading:
	    withNode doctree element HEAD child element TITLE {
		output <H1>
		output [content]
		output </H1>
	    }

	    # Build a table of contents, with one entry for each section:
	    output "<H2>Contents</H2>\n"
	    output <UL>
	    set i 1
	    foreachNode subtree element SECT {
		output "<LI><A HREF=\"#SEC${i}\">"
		withNode child element TITLE {
		    output [content] 
		}
		incr i
		output "</A>\n"
		#foreachNode child element QNA child element Q {
		#	output [content]; output " "
		#}
	    }
	    output </UL>
	}
	endAction {
	    # Add a footer:
	    output <HR>
	    output <ADDRESS>
	    withNode doctree element HEAD child element AUTHOR {
		if [query? hasatt EMAIL] {
		    output "<A HREF=\"mailto:[query attval EMAIL]\">"
		}
		output [content];	# author name
		if [query? hasatt EMAIL] {
		    output </A>
		}
	    }
	    output </ADDRESS>
	    output <I>
	    output "Revision: [query parent attval REVISION]"
	    output </I>
	}
	after	</BODY>
    }
}
