# Latex.tcl
# CoST SGML->LaTeX (2e) conversion utilities
#

needExtension ITCL

itcl_class LatexProcess {
    inherit EventHandler

    # supply either -specname <name> or -specification <spec-args>
    public specname ;
    public specification ;
    public verbose 0;
    public secheadings "section subsection subsubsection paragraph"
    public preamble ""; # extra preamble, inserted before \begin{document}

    protected spec;
    protected env;
    protected undefinedGIs "";
    protected blanklines -1;
    protected seclevel -1;

    constructor {config} {
	if [info exists specname] {
	    set spec $specname 
	} else {
	    set spec [specification ${this}Assoc $specification]
	}
	set env [environment "${this}Env" \
			cdataFilter latexEscape \
			sdataFilter sdataFilter \
			markupFilter identityFilter \
			suppress 0 \
		]
    }

#
# Output methods:
# Blank lines are significant in TeX (they cause a paragraph break),
# so these methods keep track of newlines.
# Use 'output' for character data (escaped), 'insert' for LaTeX markup.
#
    method output {text} {
	if {$text != ""} {
	    puts stdout $text nonewline
	    set blanklines -1
	}
    }
    method insert {text} {
	output [[$env get markupFilter] $text]
    }
#
# 'nl n' appends 'n' blank lines; if n==0, just break the line. 
# NB: 'output' assumes that its argument contains no newlines.
#
    method nl {{n 0}} {
	while {$blanklines < $n} {
	    puts stdout "";
	    incr blanklines
	}
    }

#
# Event handler methods:
#
    method CDATA {} { output [[$env get cdataFilter] [content]] }
    method SDATA {} { output [[$env get sdataFilter] [content]] }
    method RE {} { nl 0 }
    method DATAENT {} { specialStart }

#
# START and END methods just call the architectural form handlers,
# inserting the 'before', 'after', 'prefix' and 'suffix' parameters verbatim
#
    method START {} {
	insert [$spec get before ""]
	"[$spec get arcform undefined]Start"
	insert [$spec get prefix ""]
    }

    method END {} {
	insert [$spec get suffix ""]
	"[$spec get arcform undefined]End"
	insert [$spec get after ""]
    }

#
# LaTeX "architectural form" processing:
# New forms may be defined by subclassing the LatexProcess class,
# adding xxxStart and xxxEnd methods for each new form 'xxx'
#

    # none: no special processing
    method noneStart {} { }
    method noneEnd {} { }

    # ignore: supress content of element
    # parameters:
    #	content: replacement content; $variables and [commands] expanded
    method ignoreStart {} {
	output [subst [$spec get content ""]]
	$env save cdataFilter nullFilter \
		    sdataFilter nullFilter \
		    markupFilter nullFilter \
		    suppress 1
    }
    method ignoreEnd {} {
	$env restore
    }

    # undefined: called as a default if no form specified.
    # Prints a warning to that effect if -verbose 1 specified
    method undefinedStart {} {
	if ![$env get suppress] {
	    set gi [query gi]
	    if { [lsearch $undefinedGIs $gi] == -1 } {
		lappend undefinedGIs $gi
		if $verbose { warning "No mapping for $gi element" }
	    }
	}
    }
    method undefinedEnd {} { }
    method queryUndefined {} { return $undefinedGIs }
    method resetUndefined {} { set undefinedGIs "" }

    # special: special processing
    # parameters:
    #	content:	content of element, with Tcl $variable
    #			and [command] substitution.
    #   startAction:
    #	endAction:	Tcl code evaluated at start/end of element
    method specialStart {} {
	$spec do startAction
	output [subst [$spec get content ""]]
    }
    method specialEnd {} { 
	$spec do endAction
    }

    # environment: maps to  \begin{xxx} ... \end{xxx}
    # parameters: 
    #	envname:	name of environment
    #	envargs:	extra arguments to \begin{envname}, incl. {}[]s.
    method environmentStart {} {
	nl;
	insert "\\begin\{[$spec get envname]\}[$spec get envargs ""]"
	nl 0
    }
    method environmentEnd {} {
	nl 0; insert "\\end\{[$spec get envname]\}"; nl;
    }

    # macro: maps to \xxx{ ... }
    # parameters:
    #	macro:		control sequence name (w/o backslash)
    method macroStart {} { insert "\\[$spec get macro]\{" }
    method macroEnd {} { insert "\}" }

    # paragraph: a paragraph.
    method paragraphStart {} { nl 1 }
    method paragraphEnd {} { nl 1 }

    # document: document body
    # parameters:
    #	documentclass:	LaTeX document class
    #	classoptions:	list of class options
    #	packages:	list of packages (.sty files) used
    #
    method documentStart {} {
	set class [$spec get documentclass article]
	set options [$spec get classoptions ""]
	if {$options != ""} { set options "\[$options\]" }
	output "% Generated by CoST"; nl
	output "\\documentclass${options}{$class}"; nl
	foreach package [$spec get packages {}] {
	    output "\\usepackage{$package}"; nl
	}
	nl; output $preamble; nl;
	output [subst [$spec get preamble ""]]; nl;
	output "\\begin{document}"; nl
    }
    method documentEnd {} {
	nl; output "\\end{document}"; nl
    }
    # %%% Need: author, title, date, other metainfo

    # division: hierarchical division (section, chapter, etc.)
    # In LaTeX, sections are specified by heading elements, not containers.
    # The actual heading is produced by the 'divheading' form element
    method divisionStart {} {
	nl 1; incr seclevel;
	setprop seclevel $seclevel
	if [query? hasatt ID] { setprop reflabel [query attval ID] }
    }
    method divisionEnd {} {
	nl 1; incr seclevel -1
    }

    # divheading: hierarchical division title
    # parameters:
    #	macro:	macro name for division heading; 
    #		(default: determined by section level and $secheadings list)
    method divheadingStart {} {
	if [$spec has macro ] {
	    macroStart
	} else {
	    insert "\\[lindex $secheadings $seclevel]\{"
	}
    }
    method divheadingEnd {} {
	if [$spec has macro] {
	    macroEnd
	} else {
	    insert "\}"
	}
	# add \label{ID} if containing SECTion element has an ID: 
	if [query? parent hasprop reflabel] {
	    insert " \\label{[query parent propval reflabel]}"
	}
	nl;
    }
    # verbatim: special case of 'environment'; 
    # does not escape special characters in content.
    # parameters:
    #	cdataFilter:	content filter procedure (default: tabexFilter)
    #	envname:	environment name (default: verbatim)
    method verbatimStart {} {
	nl; insert "\\begin{[$spec get envname verbatim]}"; nl;
	# should markupFilter == nullFilter?
	$env save cdataFilter [$spec get cdataFilter tabexFilter] \
			sdataFilter nullFilter \
			markupFilter identityFilter
    }
    method verbatimEnd {} {
	$env restore
	nl; insert "\\end{[$spec get envname verbatim]}"; nl
    }

    # lists: special case of environment.
    # should only contain item, itemtag, and itembody form elements.
    method listStart {} 	{ environmentStart }
    method listEnd {} 		{ environmentEnd }

    # item: members of single-part lists like itemize and enumerate
    method itemStart {} 	{ nl; insert "\\item{}"; nl }
    method itemEnd {}  		{ nl 1 }

    # itemtag/itembody: members of two-part lists like description
    # should occur in pairs. 
    method itemtagStart {} 	{ nl; insert "\\item\[" }
    method itemtagEnd {}  	{ insert "\]" }
    method itembodyStart {} 	{ }
    method itembodyEnd {}  	{ }

    # xref: cross-references.
    # Don't ask.
    method xrefStart {} {
	set linkend [query attval REFID];
	if {$linkend == ""} { warning "XREF has no linkend"; return }
	set reflabel ""
	set refprefix ""
	withNode doctree withattval ID $linkend {
	    set refprefix [$spec get refprefix [query propval refprefix]]
	    set reftype [$spec get reftype [query propval reftype]]
	    set reflabel [query propval reflabel]
	}
	set refprefix [$spec get refprefix $refprefix]
	set reftype [$spec get reftype $reftype]
	if {$reflabel == ""} { set reflabel $linkend }
	if {$reflabel == ""} { warning "XREF to $linkend has no label" }
        insert "$refprefix\\$reftype{$reflabel}"
    }
    method xrefEnd {} { }

}

# Substitution for text mode:
# %%% This only works with T1 encoding
# %%% Trying to do this with OT1 is too horrifying to think about
substitution latexEscape {
	"\{"	"\\\{"
	"\}"	"\\\}"
	"["	"\{[\}"
	"]"	"\{]\}"
	"\\"	"\\char\"5C{}"
	"#"	"\\#"
	"$"	"\\char`$"
	"%"	"\\%"
	"&"	"\\char`&"
	"~"	"$\\sim$"
	"_"	"\\char`_"
	"^"	"\\char`^"
	"TeX"	" \\TeX{}"
	"LaTeX"	"\\LaTeX{}"
}

proc identityFilter {str} { return $str }
proc nullFilter {str} { return "" }
proc sdataFilter {str} { return $str }

# filter for 'verbatim' environment: expand tabs, 
# don't substitute anything else.
# %%% this doesn't work: subelements will throw off column count
proc tabexFilter {line} {
    if {[string first "\t" $line] == -1} {return $line}
    set spaces 0
    set tabstop 8
    set result ""
    set col 0
    foreach chunk [split $line "\t"] {
	incr col $spaces
	while {$spaces > 0} { append result " "; incr spaces -1 }
	append result $chunk
	incr col [string length $chunk]
	set spaces [expr $tabstop - $col % $tabstop]
    }
    return $result
}

