/***********************************************************************
 *               Copyright (C) 1995 Joe English
 *                   Freely redistributable
 ***********************************************************************
 *
 * File:	assoc.c
 * Author:	Joe English
 * Created:	23 Mar 95 / 10 Jun 95
 * Description:	Cost 'associations' module.
 * 		Map list of selection queries to set of name/value bindings
 *
 * BUGS:	query list processing could use some optimization.
 *		Add generalized concept of 'parameter sets',
 *		and cache matched association with node;
 * 		how to do this is unclear.
 * 		testing query lists could be optimized
 * 		with the same technique as Haskell/ML-style pattern matching
 *
 * $Date: 95/08/31 16:26:16 $
 * $Revision: 1.5 $
 * $Locker:  $
 */

#ifndef lint
static char RCSId[] = "$Id: assoc.c,v 1.5 95/08/31 16:26:16 joe Exp $";
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "tcl.h"
#include "project.h"
#include "strmgt.h"
#include "strmap.h"
#include "esis.h"
#include "costq.h"
#include "tclcost.h"

/*+++ Local definitions:
 */

struct CQAssoc
{
    CQQuery	query;
    strmap	bindings;
    CQAssocRec  *next;
};

/*+++ Associations
 */

void assocDestructor(void *closure)
{
    CQAssoc assoc = (CQAssoc)closure;
    while (assoc)
    {
	if (assoc->query) cq_destroyquery(assoc->query);
	if (assoc->bindings) strmap_destroy(assoc->bindings);
	assoc = assoc->next;
    }
}

CQAssoc assocConstructor(Tcl_Interp *interp, char *str)
{
    CQAssoc assoc = NULL;
    CQAssoc *tailptr = &assoc;
    char **assocpairs = NULL;
    int nassocpairs;
    int status;
    int i;

    status = Tcl_SplitList(interp, str, &nassocpairs, &assocpairs);
    if (status == TCL_ERROR) return 0;

    if (nassocpairs & 1) {
	Tcl_SetErrorCode(interp,"COST","ASSOC","Odd number of assocations",0);
	goto err;
    }

    for (i=0; i<nassocpairs; i+=2)
    {
	CQAssocRec *a;
	char **sublist;
	int sublistlen,j;

	/* First pair: query */
	status = Tcl_SplitList(interp, assocpairs[i], &sublistlen, &sublist);
	if (status == TCL_ERROR) goto err;

	a = malloc(sizeof(*a));
	a->query = cq_buildquery(sublist, sublistlen);
	a->bindings = strmap_create();
	a->next = 0;
	*tailptr = a;
	tailptr = &a->next;
	free(sublist);
	if (!a->query)
		goto err;

	status = Tcl_SplitList(interp, assocpairs[i+1], &sublistlen, &sublist);
	if (status == TCL_ERROR) goto err;

	if (sublistlen % 2 != 0) goto err;
	for (j=0; j < sublistlen; j += 2)
	    strmap_set(a->bindings, sublist[j], sublist[j+1]);

	free(sublist);
    }

    free(assocpairs);
    return assoc;

err:
    if (assocpairs) free(assocpairs);
    assocDestructor(assoc);
    return 0;
}

static char *assocLookup(CQAssoc assoc, ESISNode node, const char *name)
{
    while (assoc)
    {
	if (cq_testquery(node, assoc->query)) 
	{
	    char *value = strmap_get(assoc->bindings, name);
	    if (value)
		return value;
	}
	assoc = assoc->next;
    }
    return 0;
}

int assocProc(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
    CQAssoc assoc = (CQAssoc)clientData;
    char *subcmd = argv[1];

    if (argc <= 1) goto usage;
    CHECKCURNODE;

    if (isupper(subcmd[0]))
    { /* Assume it's an event handler/method name -- lookup and execute */
      /* %%% This is either a cheap hack or an undocumented feature,
       * %%% depending on how you look at it...
       */
	char *script = assocLookup(assoc, current_node, subcmd);
	if (script)
	    return Tcl_Eval(interp, script);
	/* Calling non-existant handler is a no-op */
	return TCL_OK;
    }
    else if (!strcmp(subcmd,"get"))
    {
	char *retval = 0;
	if (argc < 3 || argc > 4) goto usage;
	retval = assocLookup(assoc, current_node, argv[2]);
	if (!retval) {
	    if (argc >= 4)  {
		retval = argv[3];
	    } else {
		Tcl_AppendResult(interp,
			argv[0], ": no binding for ", argv[2], NULL);
		return TCL_ERROR;
	    }
	}
	Tcl_SetResult(interp, retval, TCL_VOLATILE);
	return TCL_OK;
    }
    else if (!strcmp(subcmd,"has"))
    {
	if (argc != 3) goto usage;
	if (assocLookup(assoc, current_node, argv[2]) != 0)
	    Tcl_SetResult(interp,"1",TCL_STATIC);
	else
	    Tcl_SetResult(interp,"0",TCL_STATIC);
	return TCL_OK;
    }
    else if (!strcmp(subcmd,"do"))
    {
	char *script = 0;
	if (argc != 3) goto usage;
	script = assocLookup(assoc, current_node, argv[2]);
	if (script)
	    return Tcl_Eval(interp, script);
	return TCL_OK;
    }

    /* else -- fallthru to usage */
usage:
    Tcl_AppendResult(interp, "Usage: ",
	argv[0],
	" get name ?default?",
	" | has name ",
	" | do method",	/* don't mention START/END/etc */
	0);

    return TCL_ERROR;
}
