/***********************************************************************
 *               Copyright (C) 1995 Joe English
 *                   Freely redistributable
 ***********************************************************************
 *
 * File:	relation.c
 * Author:	Joe English
 * Created: 	25 Oct 1995
 * Description:	Implementation of ILINKs and RELATIONS.
 *
 * $Date: 96/01/16 12:45:39 $
 * $Revision: 1.4 $
 * $Locker:  $
 */

#ifndef lint
static char RCSId[] = "$Id: relation.c,v 1.4 96/01/16 12:45:39 joe Exp $";
#endif

#include <stdlib.h>
#include <stdio.h>

#include "project.h"
#include "strmgt.h"
#include "esis.h"
#include "strmap.h"
#include "pile.h"
#include "esisp.h"


/* Organization:
 *
 * ILINK node structure:
 *	name:		relation name
 *	parent:		origin of link
 *	children:	LINKEND nodes
 *	next,prev:	next/previous ilink in relation
 *
 * LINKEND node structure:
 *	name:		anchor name
 *	parent: 	containing ILINK node
 *	reference:	anchor node 
 *	next,prev:	next/previous anchor in relation
 *
 * All the LINKEND nodes which reference a particular tree node
 * are stored in a circular linked list, chained by the 'link' 
 * field.   The 'link' field of the tree node points to the tail of 
 * this list.
 */

struct ESISRelation 
{
    ESISToken	name;		/* relation name */
    struct ESISRelation *next; 	/* next relation */
    ESISNode	head;		/* first ILINK node in relation */
    ESISNode	tail;		/* last ILINK node in relation */
    int		nanchors;	/* #anchors */
    ESISToken	*anchnames;	/* array of anchor names */
};

/* 
 * Constructors:
 */
static ESISRelation find_relation(ESISStream ep, ESISToken name)
{
    ESISRelation rel = ep->relations;

    while (rel && rel->name != name)
	rel=rel->next;
    return rel;
}

static ESISNode find_linkend(ESISNode ilink, ESISToken anchname)
{
    ESISNode linkend = ilink->children;
    while (linkend) 
    {
	if (linkend->name == anchname)
	    break;
	linkend = linkend->next;
    }
    return linkend;
}

int esis_define_relation(
	ESISStream ep, 
	const char *relname,
	int nanchors, 
	char **anchnames)
{
    ESISRelation rel;
    ESISToken name = ucintern(relname);
    int i;

    if (find_relation(ep, name))
	return 0;

    rel = malloc(sizeof(*rel));
    rel->next = ep->relations;	ep->relations = rel;
    rel->name = name;
    rel->head = 0;
    rel->tail = 0;
    rel->nanchors = nanchors;
    rel->anchnames = malloc(nanchors * sizeof(ESISToken));
    for (i=0; i<nanchors; ++i) {
	rel->anchnames[i] = ucintern(anchnames[i]);
    }
    return 1;
}

ESISNode esis_create_ilink(ESISStream ep, ESISToken relname, ESISNode origin)
{
    ESISRelation rel;
    ESISNode ilink,linkend;
    int i;

    rel = find_relation(ep,relname);
    if (!rel)
	return 0;
    
    ilink = esis_create_node(EN_ILINK, 0, 0);
    ilink->name = relname;
    ilink->parent = origin;	/* ??? should be reference? */

    if (!rel->head) 
	rel->head = ilink;
    else 
	rel->tail->next = ilink;
    rel->tail = ilink;

    for (i=0,linkend=NULL; i<rel->nanchors; ++i) {
	linkend = esis_create_node(EN_LINKEND, ilink, linkend);
	linkend->name = rel->anchnames[i];
    }

    return ilink;
}

int esis_set_linkend(ESISNode ilink, ESISToken anchname, ESISNode anchor)
{
    ESISNode linkend = find_linkend(ilink, anchname);

    if (!linkend)
	return 0;
    if (linkend->reference)
	return 0;
    linkend->reference = anchor;

    /* Add to circular linked list: */
    if (anchor->link)
    {
	linkend->link = anchor->link->link;
	anchor->link->link = linkend;
	anchor->link = linkend;
    } else
    {
	anchor->link = linkend->link = linkend;
    }

    return 1;
}

ESISNode esis_relation_first(ESISStream ep, ESISToken relname)
{
    ESISRelation rel = find_relation(ep, relname);
    return rel ? rel->head : 0; 
}

ESISNode esis_relation_next(ESISNode ilink)
{
    if (ilink->type != EN_ILINK)
	return 0;
    return ilink->next;
}

ESISNode esis_first_ilink(
	ESISNode origin,
	ESISToken relname,
	ESISToken anchname)
{
    ESISNode linkend = origin->link;
    ASSERT(relname == ucintern(relname), "Uninterned relname");
    ASSERT(anchname == ucintern(anchname), "Uninterned anchname");

    if (!linkend) return 0;

    linkend = linkend->link;	/* head of cll */
    do {
	ASSERT(linkend->type == EN_LINKEND 
	   && linkend->parent->type == EN_ILINK, 
	   "Botched ilink structure");
	if (linkend->name == anchname && linkend->parent->name == relname)
	    return linkend->parent;
	linkend = linkend->link;
    } while (linkend != origin->link);
    return 0;
}

ESISNode esis_next_ilink(
	ESISNode origin,
	ESISToken relname,
	ESISToken anchname,
	ESISNode ilink)
{
    ESISNode linkend;

    ASSERT(ilink->type == EN_ILINK, "Not an ILINK node");
    linkend = find_linkend(ilink, anchname);
    while (linkend != origin->link) 
    {
	linkend = linkend->link;
	ASSERT(linkend->type == EN_LINKEND 
	   && linkend->parent->type == EN_ILINK, 
	   "Botched ilink structure");
	if (linkend->name == anchname && linkend->parent->name == relname)
	    return linkend->parent;
    }
    return 0;
}

ESISNode esis_ilink_anchor(ESISNode ilink, ESISToken anchname)
{
    ESISNode linkend;
    if (ilink->type != EN_ILINK)
	return 0;
    linkend = find_linkend(ilink, anchname);
    return linkend ? linkend->reference : 0;
}

ESISNode esis_ilink_origin(ESISNode ilink)
{
    if (ilink->type == EN_ILINK)
	return ilink->parent;
    return NULL;
}
